@@ -18,167 +18,174 @@ from django.db.models import Case, When
 from django.core.exceptions import ValidationError
 
 LIMITING_FK_EDGES_CLAUSE_1 = (
-    """AND second.{fk_field_name}_id = %(limiting_fk_edges_instance_id)s"""
+    """AND second.{fk_field_name}_{pk_name} = %(limiting_fk_edges_instance_pk)s"""
 )
-LIMITING_FK_EDGES_CLAUSE_2 = """AND {relationship_table}.{fk_field_name}_id = %(limiting_fk_edges_instance_id)s"""
+LIMITING_FK_EDGES_CLAUSE_2 = """AND {relationship_table}.{fk_field_name}_{pk_name} = %(limiting_fk_edges_instance_pk)s"""
 
 LIMITING_FK_NODES_CLAUSE_1 = """"""
 LIMITING_FK_NODES_CLAUSE_2 = """"""
 
-EXCLUDED_UPWARD_NODES_CLAUSE_1 = """AND second.child_id <> ALL(%(excluded_upward_node_ids)s::int[])"""  # Used for ancestors and upward path
-EXCLUDED_UPWARD_NODES_CLAUSE_2 = (
-    """AND {relationship_table}.child_id <> ALL(%(excluded_upward_node_ids)s::int[])"""
-)
+# DISALLOWED_ANCESTORS_NODES_CLAUSE_1 = """AND second.child_pk <> ALL(%(disallowed_ancestors_node_pks)s)"""  # Used for ancestors and upward path
+# DISALLOWED_ANCESTORS_NODES_CLAUSE_2 = ("""AND {relationship_table}.child_pk <> ALL(%(disallowed_ancestors_node_pks)s)""")
 
-EXCLUDED_DOWNWARD_NODES_CLAUSE_1 = """AND second.parent_id <> ALL(%(excluded_downward_node_ids)s::int[])"""  # Used for descendants and downward path
-EXCLUDED_DOWNWARD_NODES_CLAUSE_2 = """AND {relationship_table}.parent_id <> ALL(%(excluded_downward_node_ids)s::int[])"""
+# DISALLOWED_DESCENDANTS_NODES_CLAUSE_1 = """AND second.parent_pk <> ALL(%(disallowed_descendants_node_pks)s)"""  # Used for descendants and downward path
+# DISALLOWED_DESCENDANTS_NODES_CLAUSE_2 = """AND {relationship_table}.parent_pk <> ALL(%(disallowed_descendants_node_pks)s)"""
 
-REQUIRED_UPWARD_NODES_CLAUSE_1 = """"""  # Used for ancestors and upward path
-REQUIRED_UPWARD_NODES_CLAUSE_2 = """"""
+DISALLOWED_ANCESTORS_NODES_CLAUSE_1 = """AND first.parent_{pk_name} <> ALL(%(disallowed_ancestors_node_pks)s)"""  # Used for ancestors and upward path
+DISALLOWED_ANCESTORS_NODES_CLAUSE_2 = """AND {relationship_table}.parent_{pk_name} <> ALL(%(disallowed_ancestors_node_pks)s)"""
 
-REQUIRED_DOWNWARD_NODES_CLAUSE_1 = """"""  # Used for descendants and downward path
-REQUIRED_DOWNWARD_NODES_CLAUSE_2 = """"""
+DISALLOWED_DESCENDANTS_NODES_CLAUSE_1 = """AND first.child_{pk_name} <> ALL(%(disallowed_descendants_node_pks)s)"""  # Used for descendants and downward path
+DISALLOWED_DESCENDANTS_NODES_CLAUSE_2 = """AND {relationship_table}.child_{pk_name} <> ALL(%(disallowed_descendants_node_pks)s)"""
+
+
+ALLOWED_ANCESTORS_NODES_CLAUSE_1 = """AND first.parent_pk = ANY(%(allowed_ancestors_node_pks)s)"""  # Used for ancestors and upward path
+ALLOWED_ANCESTORS_NODES_CLAUSE_2 = """AND {relationship_table}.parent_{pk_name} = ANY(%(allowed_ancestors_node_pks)s)"""
+
+ALLOWED_DESCENDANTS_NODES_CLAUSE_1 = """AND first.child_{pk_name} = ANY(%(allowed_descendants_node_pks)s)"""  # Used for descendants and downward path
+ALLOWED_DESCENDANTS_NODES_CLAUSE_2 = """AND {relationship_table}.child_{pk_name} = ANY(%(allowed_descendants_node_pks)s)"""
 
 ANCESTORS_QUERY = """
-WITH RECURSIVE traverse(id, depth) AS (
-    SELECT first.parent_id, 1
+WITH RECURSIVE traverse({pk_name}, depth) AS (
+    SELECT first.parent_{pk_name}, 1
         FROM {relationship_table} AS first
         LEFT OUTER JOIN {relationship_table} AS second
-        ON first.parent_id = second.child_id
-    WHERE first.child_id = %(id)s
+        ON first.parent_{pk_name} = second.child_{pk_name}
+    WHERE first.child_{pk_name} = %(pk)s
     -- LIMITING_FK_EDGES_CLAUSE_1
-    -- EXCLUDED_UPWARD_NODES_CLAUSE_1
-    -- REQUIRED_UPWARD_NODES_CLAUSE_1
+    -- DISALLOWED_ANCESTORS_NODES_CLAUSE_1
+    -- ALLOWED_ANCESTORS_NODES_CLAUSE_1
     {ancestors_clauses_1}
 UNION
-    SELECT DISTINCT parent_id, traverse.depth + 1
+    SELECT DISTINCT parent_{pk_name}, traverse.depth + 1
         FROM traverse
         INNER JOIN {relationship_table}
-        ON {relationship_table}.child_id = traverse.id
+        ON {relationship_table}.child_{pk_name} = traverse.{pk_name}
     WHERE 1 = 1
     -- LIMITING_FK_EDGES_CLAUSE_2
-    -- EXCLUDED_UPWARD_NODES_CLAUSE_2
-    -- REQUIRED_UPWARD_NODES_CLAUSE_2
+    -- DISALLOWED_ANCESTORS_NODES_CLAUSE_2
+    -- ALLOWED_ANCESTORS_NODES_CLAUSE_2
     {ancestors_clauses_2}
 )
-SELECT id FROM traverse
-GROUP BY id
-ORDER BY MAX(depth) DESC, id ASC
+SELECT {pk_name} FROM traverse
+WHERE depth <= %(max_depth)s
+GROUP BY {pk_name}
+ORDER BY MAX(depth) DESC, {pk_name} ASC
 """
 
 DESCENDANTS_QUERY = """
-WITH RECURSIVE traverse(id, depth) AS (
-    SELECT first.child_id, 1
+WITH RECURSIVE traverse({pk_name}, depth) AS (
+    SELECT first.child_{pk_name}, 1
         FROM {relationship_table} AS first
         LEFT OUTER JOIN {relationship_table} AS second
-        ON first.child_id = second.parent_id
-    WHERE first.parent_id = %(id)s
+        ON first.child_{pk_name} = second.parent_{pk_name}
+    WHERE first.parent_{pk_name} = %(pk)s
     -- LIMITING_FK_EDGES_CLAUSE_1
-    -- EXCLUDED_DOWNWARD_NODES_CLAUSE_1
-    -- REQUIRED_DOWNWARD_NODES_CLAUSE_1
+    -- DISALLOWED_DESCENDANTS_NODES_CLAUSE_1
+    -- ALLOWED_DESCENDANTS_NODES_CLAUSE_1
     {descendants_clauses_1}
 UNION
-    SELECT DISTINCT child_id, traverse.depth + 1
+    SELECT DISTINCT child_{pk_name}, traverse.depth + 1
         FROM traverse
         INNER JOIN {relationship_table}
-        ON {relationship_table}.parent_id = traverse.id
+        ON {relationship_table}.parent_{pk_name} = traverse.{pk_name}
     WHERE 1=1
     -- LIMITING_FK_EDGES_CLAUSE_2
-    -- EXCLUDED_DOWNWARD_NODES_CLAUSE_2
-    -- REQUIRED_DOWNWARD_NODES_CLAUSE_2
+    -- DISALLOWED_DESCENDANTS_NODES_CLAUSE_2
+    -- ALLOWED_DESCENDANTS_NODES_CLAUSE_2
     {descendants_clauses_1}
 )
-SELECT id FROM traverse
-GROUP BY id
-ORDER BY MAX(depth), id ASC
+SELECT {pk_name} FROM traverse
+WHERE depth <= %(max_depth)s
+GROUP BY {pk_name}
+ORDER BY MAX(depth), {pk_name} ASC
 """
 
 PATH_LIMITING_FK_EDGES_CLAUSE = (
-    """AND first.{fk_field_name}_id = %(limiting_fk_edges_instance_id)s"""
+    """AND first.{fk_field_name}_{pk_name} = %(limiting_fk_edges_instance_pk)s"""
 )
 PATH_LIMITING_FK_NODES_CLAUSE = """"""
 
-EXCLUDED_UPWARD_PATH_NODES_CLAUSE = (
-    """AND second.parent_id <> ALL('{excluded_path_node_ids}'::int[])"""
+DISALLOWED_UPWARD_PATH_NODES_CLAUSE = (
+    """AND second.parent_{pk_name} <> ALL('{disallowed_path_node_pks}')"""
 )
-EXCLUDED_DOWNWARD_PATH_NODES_CLAUSE = (
-    """AND second.child_id <> ALL('{excluded_path_node_ids}'::int[])"""
+DISALLOWED_DOWNWARD_PATH_NODES_CLAUSE = (
+    """AND second.child_{pk_name} <> ALL('{disallowed_path_node_pks}')"""
 )
-REQUIRED_UPWARD_PATH_NODES_CLAUSE = (
-    """AND second.parent_id = ALL('{required_path_node_ids}'::int[])"""
+ALLOWED_UPWARD_PATH_NODES_CLAUSE = (
+    """AND second.parent_{pk_name} = ALL('{allowed_path_node_pks}')"""
 )
-REQUIRED_DOWNWARD_PATH_NODES_CLAUSE = (
-    """AND second.child_id = ALL('{required_path_node_ids}'::int[])"""
+ALLOWED_DOWNWARD_PATH_NODES_CLAUSE = (
+    """AND second.child_{pk_name} = ALL('{allowed_path_node_pks}')"""
 )
 
 UPWARD_PATH_QUERY = """
-WITH RECURSIVE traverse(child_id, parent_id, depth, path) AS (
+WITH RECURSIVE traverse(child_{pk_name}, parent_{pk_name}, depth, path) AS (
     SELECT
-        first.child_id,
-        first.parent_id,
+        first.child_{pk_name},
+        first.parent_{pk_name},
         1 AS depth,
-        ARRAY[first.child_id] AS path
+        ARRAY[first.child_{pk_name}] AS path
         FROM {relationship_table} AS first
-    WHERE child_id = %(starting_node)s
+    WHERE child_{pk_name} = %(starting_node)s
 UNION ALL
     SELECT
-        first.child_id,
-        first.parent_id,
+        first.child_{pk_name},
+        first.parent_{pk_name},
         second.depth + 1 AS depth,
-        path || first.child_id AS path
+        path || first.child_{pk_name} AS path
         FROM {relationship_table} AS first, traverse AS second
-    WHERE first.child_id = second.parent_id
-    AND (first.child_id <> ALL(second.path))
+    WHERE first.child_{pk_name} = second.parent_{pk_name}
+    AND (first.child_{pk_name} <> ALL(second.path))
     -- PATH_LIMITING_FK_EDGES_CLAUSE
-    -- EXCLUDED_UPWARD_PATH_NODES_CLAUSE
-    -- REQUIRED_UPWARD_PATH_NODES_CLAUSE
+    -- DISALLOWED_UPWARD_PATH_NODES_CLAUSE
+    -- ALLOWED_UPWARD_PATH_NODES_CLAUSE
     -- LIMITING_UPWARD_NODES_CLAUSE_1  -- CORRECT?
     {upward_clauses}
-    AND second.depth <= %(max_depth)s
 )
 SELECT 
-    UNNEST(ARRAY[id]) AS id
+    UNNEST(ARRAY[{pk_name}]) AS {pk_name}
 FROM 
     (
-    SELECT path || ARRAY[%(ending_node)s] FROM traverse
-        WHERE parent_id = %(ending_node)s
+    SELECT path || ARRAY[%(ending_node)s], depth FROM traverse
+        WHERE parent_{pk_name} = %(ending_node)s
+        AND depth <= %(max_depth)s
         LIMIT 1
-) AS x(id);
+) AS x({pk_name});
 """
 
 DOWNWARD_PATH_QUERY = """
-WITH RECURSIVE traverse(parent_id, child_id, depth, path) AS (
+WITH RECURSIVE traverse(parent_{pk_name}, child_{pk_name}, depth, path) AS (
     SELECT
-        first.parent_id,
-        first.child_id,
+        first.parent_{pk_name},
+        first.child_{pk_name},
         1 AS depth,
-        ARRAY[first.parent_id] AS path
+        ARRAY[first.parent_{pk_name}] AS path
         FROM {relationship_table} AS first
-    WHERE parent_id = %(starting_node)s
+    WHERE parent_{pk_name} = %(starting_node)s
 UNION ALL
     SELECT
-        first.parent_id,
-        first.child_id,
+        first.parent_{pk_name},
+        first.child_{pk_name},
         second.depth + 1 AS depth,
-        path || first.parent_id AS path
+        path || first.parent_{pk_name} AS path
         FROM {relationship_table} AS first, traverse AS second
-    WHERE first.parent_id = second.child_id
-    AND (first.parent_id <> ALL(second.path))
+    WHERE first.parent_{pk_name} = second.child_{pk_name}
+    AND (first.parent_{pk_name} <> ALL(second.path))
     -- PATH_LIMITING_FK_EDGES_CLAUSE
-    -- EXCLUDED_DOWNWARD_PATH_NODES_CLAUSE
-    -- REQUIRED_DOWNWARD_PATH_NODES_CLAUSE
+    -- DISALLOWED_DOWNWARD_PATH_NODES_CLAUSE
+    -- ALLOWED_DOWNWARD_PATH_NODES_CLAUSE
     -- LIMITING_DOWNWARD_NODES_CLAUSE_1  -- CORRECT?
     {downward_clauses}
-    AND second.depth <= %(max_depth)s
 )      
 SELECT 
-    UNNEST(ARRAY[id]) AS id
+    UNNEST(ARRAY[{pk_name}]) AS {pk_name}
 FROM 
     (
-    SELECT path || ARRAY[%(ending_node)s] FROM traverse
-        WHERE child_id = %(ending_node)s
+    SELECT path || ARRAY[%(ending_node)s], depth FROM traverse
+        WHERE child_{pk_name} = %(ending_node)s
+        AND depth <= %(max_depth)s
         LIMIT 1
-) AS x(id);
+) AS x({pk_name});
 """
 
 
@@ -196,8 +203,8 @@ def _filter_order(queryset, field_names, values):
     orders results in the same order as provided values
 
         For instance
-            _filter_order(self.__class__.objects, "pk", ids)
-        returns a queryset of the current class, with instances where the 'pk' field matches an id in ids
+            _filter_order(self.__class__.objects, "pk", pks)
+        returns a queryset of the current class, with instances where the 'pk' field matches an pk in pks
 
     """
     if not isinstance(field_names, list):
@@ -217,7 +224,7 @@ def node_factory(edge_model, children_null=True, base_model=models.Model):
     def get_foreign_key_field(instance=None):
         """
         Provided a model instance and model class, checks if the edge model has a ForeignKey
-        field to the model for that instance, and then returns the field name and instance id.
+        field to the model for that instance, and then returns the field name and instance pk.
         """
         if instance is not None:
             for field in edge_model._meta.get_fields():
@@ -238,37 +245,59 @@ def node_factory(edge_model, children_null=True, base_model=models.Model):
         class Meta:
             abstract = True
 
-        def add_child(self, descendant, **kwargs):
-            kwargs.update({"parent": self, "child": descendant})
+        def add_child(self, child, **kwargs):
+            kwargs.update({"parent": self, "child": child})
             disable_check = kwargs.pop("disable_circular_check", False)
             cls = self.children.through(**kwargs)
             return cls.save(disable_circular_check=disable_check)
 
-        def remove_child(self, descendant):
-            self.children.through.objects.get(parent=self, child=descendant).delete()
+        def remove_child(self, child, delete_node=False):
+            """Removes the edge connecting this node to child, and optionally deletes the child node as well"""
+            if child in self.children.all():
+                self.children.through.objects.get(parent=self, child=child).delete()
+                if delete_node:
+                    # Note: Per django docs:
+                    # https://docs.djangoproject.com/en/dev/ref/models/instances/#deleting-objects
+                    # This only deletes the object in the database; the Python instance will still 
+                    # exist and will still have data in its fields.
+                    child.delete()
 
         def add_parent(self, parent, *args, **kwargs):
             return parent.add_child(self, **kwargs)
 
-        def remove_parent(self, parent):
-            parent.children.through.objects.get(parent=parent, child=self).delete()
+        def remove_parent(self, parent, delete_node=False):
+            """Removes the edge connecting this node to parent, and optionally deletes the parent node as well"""
+            if parent in self.parents.all():
+                parent.children.through.objects.get(parent=parent, child=self).delete()
+                if delete_node:
+                    # Note: Per django docs:
+                    # https://docs.djangoproject.com/en/dev/ref/models/instances/#deleting-objects
+                    # This only deletes the object in the database; the Python instance will still 
+                    # exist and will still have data in its fields.
+                    parent.delete()
 
-        def filter_order_ids(self, ids):
+        def filter_order_pks(self, pks):
             """
-            Generates a queryset, based on the current class and the provided ids
+            Generates a queryset, based on the current class and the provided pks
             """
-            return _filter_order(self.__class__.objects, "pk", ids)
+            return _filter_order(self.__class__.objects, "pk", pks)
+        
+        def get_pk_name(self):
+            """Sometimes we set a field other than 'pk' for the primary key.
+            This method is used to get the correct primary key field name for the
+            model so that raw queries return the correct information."""
+            return self._meta.pk.name
 
-        def ancestors_raw(self, **kwargs):
+        def ancestors_raw(self, max_depth=20, **kwargs):
             ancestors_clauses_1, ancestors_clauses_2 = ("", "")
-            query_parameters = {"id": self.id}
+            query_parameters = {"pk": self.pk, "max_depth": max_depth}
 
             limiting_fk_nodes_instance = kwargs.get("limiting_fk_nodes_instance", None)
             limiting_fk_edges_instance = kwargs.get("limiting_fk_edges_instance", None)
-            excluded_nodes_queryset = kwargs.get("excluded_nodes_queryset", None)
-            excluded_edges_queryset = kwargs.get("excluded_edges_queryset", None)
-            required_nodes_queryset = kwargs.get("required_nodes_queryset", None)
-            required_edges_queryset = kwargs.get("required_edges_queryset", None)
+            disallowed_nodes_queryset = kwargs.get("disallowed_nodes_queryset", None)
+            disallowed_edges_queryset = kwargs.get("disallowed_edges_queryset", None)
+            allowed_nodes_queryset = kwargs.get("allowed_nodes_queryset", None)
+            allowed_edges_queryset = kwargs.get("allowed_edges_queryset", None)
 
             if limiting_fk_nodes_instance is not None:
                 pass  # Not implemented yet
@@ -281,41 +310,57 @@ def node_factory(edge_model, children_null=True, base_model=models.Model):
                 if fk_field_name is not None:
                     ancestors_clauses_1 += "\n" + LIMITING_FK_EDGES_CLAUSE_1.format(
                         relationship_table=edge_model_table,
+                    pk_name=self.get_pk_name(),
                         fk_field_name=fk_field_name,
                     )
                     ancestors_clauses_2 += "\n" + LIMITING_FK_EDGES_CLAUSE_2.format(
                         relationship_table=edge_model_table,
+                    pk_name=self.get_pk_name(),
                         fk_field_name=fk_field_name,
                     )
                     query_parameters[
-                        "limiting_fk_edges_instance_id"
-                    ] = limiting_fk_edges_instance.id
+                        "limiting_fk_edges_instance_pk"
+                    ] = limiting_fk_edges_instance.pk
 
-            if excluded_nodes_queryset is not None:
-                ancestors_clauses_1 += "\n" + EXCLUDED_UPWARD_NODES_CLAUSE_1.format(
+            # Nodes that MUST NOT be included in the results
+            if disallowed_nodes_queryset is not None:
+                ancestors_clauses_1 += "\n" + DISALLOWED_ANCESTORS_NODES_CLAUSE_1.format(
                     relationship_table=edge_model_table,
+                    pk_name=self.get_pk_name(),
                 )
-                ancestors_clauses_2 += "\n" + EXCLUDED_UPWARD_NODES_CLAUSE_2.format(
+                ancestors_clauses_2 += "\n" + DISALLOWED_ANCESTORS_NODES_CLAUSE_2.format(
                     relationship_table=edge_model_table,
+                    pk_name=self.get_pk_name(),
                 )
-                query_parameters["excluded_upward_node_ids"] = str(
-                    set(excluded_nodes_queryset.values_list("id", flat=True))
+                query_parameters["disallowed_ancestors_node_pks"] = str(
+                    set(disallowed_nodes_queryset.values_list("pk", flat=True))
                 )
 
-            if excluded_edges_queryset is not None:
+            if disallowed_edges_queryset is not None:
                 pass  # Not implemented yet
 
-            if required_nodes_queryset is not None:
-                pass  # Not implemented yet
+            # Nodes that MAY be included in the results
+            if allowed_nodes_queryset is not None:
+                ancestors_clauses_1 += "\n" + ALLOWED_ANCESTORS_NODES_CLAUSE_1.format(
+                    relationship_table=edge_model_table,
+                    pk_name=self.get_pk_name(),
+                )
+                ancestors_clauses_2 += "\n" + ALLOWED_ANCESTORS_NODES_CLAUSE_2.format(
+                    relationship_table=edge_model_table,
+                    pk_name=self.get_pk_name(),
+                )
+                query_parameters["allowed_ancestors_node_pks"] = str(
+                    set(allowed_nodes_queryset.values_list("pk", flat=True))
+                )
 
-            if required_edges_queryset is not None:
+            if allowed_edges_queryset is not None:
                 pass  # Not implemented yet
 
             NodeModel = self._meta.model
-
             raw_qs = NodeModel.objects.raw(
                 ANCESTORS_QUERY.format(
                     relationship_table=edge_model_table,
+                    pk_name=self.get_pk_name(),
                     ancestors_clauses_1=ancestors_clauses_1,
                     ancestors_clauses_2=ancestors_clauses_2,
                 ),
@@ -324,27 +369,31 @@ def node_factory(edge_model, children_null=True, base_model=models.Model):
             return raw_qs
 
         def ancestors(self, **kwargs):
-            ids = [item.id for item in self.ancestors_raw(**kwargs)]
-            return self.filter_order_ids(ids)
+            pks = [item.pk for item in self.ancestors_raw(**kwargs)]
+            return self.filter_order_pks(pks)
+
+        def ancestors_count(self):
+            # ToDo: Implement
+            pass
 
         def self_and_ancestors(self, **kwargs):
-            ids = [self.id] + [item.id for item in self.ancestors_raw(**kwargs)][::-1]
-            return self.filter_order_ids(ids)
+            pks = [self.pk] + [item.pk for item in self.ancestors_raw(**kwargs)][::-1]
+            return self.filter_order_pks(pks)
 
         def ancestors_and_self(self, **kwargs):
-            ids = [item.id for item in self.ancestors_raw(**kwargs)] + [self.id]
-            return self.filter_order_ids(ids)
+            pks = [item.pk for item in self.ancestors_raw(**kwargs)] + [self.pk]
+            return self.filter_order_pks(pks)
 
-        def descendants_raw(self, **kwargs):
+        def descendants_raw(self, max_depth=20, **kwargs):
             descendants_clauses_1, descendants_clauses_2 = ("", "")
-            query_parameters = {"id": self.id}
+            query_parameters = {"pk": self.pk, "max_depth": max_depth}
 
             limiting_fk_nodes_instance = kwargs.get("limiting_fk_nodes_instance", None)
             limiting_fk_edges_instance = kwargs.get("limiting_fk_edges_instance", None)
-            excluded_nodes_queryset = kwargs.get("excluded_nodes_queryset", None)
-            excluded_edges_queryset = kwargs.get("excluded_edges_queryset", None)
-            required_nodes_queryset = kwargs.get("required_nodes_queryset", None)
-            required_edges_queryset = kwargs.get("required_edges_queryset", None)
+            disallowed_nodes_queryset = kwargs.get("disallowed_nodes_queryset", None)
+            disallowed_edges_queryset = kwargs.get("disallowed_edges_queryset", None)
+            allowed_nodes_queryset = kwargs.get("allowed_nodes_queryset", None)
+            allowed_edges_queryset = kwargs.get("allowed_edges_queryset", None)
 
             if limiting_fk_nodes_instance is not None:
                 pass  # Not implemented yet
@@ -357,34 +406,50 @@ def node_factory(edge_model, children_null=True, base_model=models.Model):
                 if fk_field_name is not None:
                     descendants_clauses_1 += "\n" + LIMITING_FK_EDGES_CLAUSE_1.format(
                         relationship_table=edge_model_table,
+                    pk_name=self.get_pk_name(),
                         fk_field_name=fk_field_name,
                     )
                     descendants_clauses_2 += "\n" + LIMITING_FK_EDGES_CLAUSE_2.format(
                         relationship_table=edge_model_table,
+                    pk_name=self.get_pk_name(),
                         fk_field_name=fk_field_name,
                     )
                     query_parameters[
-                        "limiting_fk_edges_instance_id"
-                    ] = limiting_fk_edges_instance.id
+                        "limiting_fk_edges_instance_pk"
+                    ] = limiting_fk_edges_instance.pk
 
-            if excluded_nodes_queryset is not None:
-                descendants_clauses_1 += "\n" + EXCLUDED_DOWNWARD_NODES_CLAUSE_1.format(
+            # Nodes that MUST NOT be included in the results
+            if disallowed_nodes_queryset is not None:
+                descendants_clauses_1 += "\n" + DISALLOWED_DESCENDANTS_NODES_CLAUSE_1.format(
                     relationship_table=edge_model_table,
+                    pk_name=self.get_pk_name(),
                 )
-                descendants_clauses_2 += "\n" + EXCLUDED_DOWNWARD_NODES_CLAUSE_2.format(
+                descendants_clauses_2 += "\n" + DISALLOWED_DESCENDANTS_NODES_CLAUSE_2.format(
                     relationship_table=edge_model_table,
+                    pk_name=self.get_pk_name(),
                 )
-                query_parameters["excluded_downward_node_ids"] = str(
-                    set(excluded_nodes_queryset.values_list("id", flat=True))
+                query_parameters["disallowed_downward_node_pks"] = str(
+                    set(disallowed_nodes_queryset.values_list("pk", flat=True))
                 )
 
-            if excluded_edges_queryset is not None:
+            if disallowed_edges_queryset is not None:
                 pass  # Not implemented yet
 
-            if required_nodes_queryset is not None:
-                pass  # Not implemented yet
+            # Nodes that MAY be included in the results
+            if allowed_nodes_queryset is not None:
+                descendants_clauses_1 += "\n" + ALLOWED_DESCENDANTS_NODES_CLAUSE_1.format(
+                    relationship_table=edge_model_table,
+                    pk_name=self.get_pk_name(),
+                )
+                descendants_clauses_2 += "\n" + ALLOWED_DESCENDANTS_NODES_CLAUSE_2.format(
+                    relationship_table=edge_model_table,
+                    pk_name=self.get_pk_name(),
+                )
+                query_parameters["allowed_descendants_node_pks"] = str(
+                    set(allowed_nodes_queryset.values_list("pk", flat=True))
+                )
 
-            if required_edges_queryset is not None:
+            if allowed_edges_queryset is not None:
                 pass  # Not implemented yet
 
             NodeModel = self._meta.model
@@ -392,6 +457,7 @@ def node_factory(edge_model, children_null=True, base_model=models.Model):
             raw_qs = NodeModel.objects.raw(
                 DESCENDANTS_QUERY.format(
                     relationship_table=edge_model_table,
+                    pk_name=self.get_pk_name(),
                     descendants_clauses_1=descendants_clauses_1,
                     descendants_clauses_2=descendants_clauses_2,
                 ),
@@ -400,27 +466,51 @@ def node_factory(edge_model, children_null=True, base_model=models.Model):
             return raw_qs
 
         def descendants(self, **kwargs):
-            ids = [item.id for item in self.descendants_raw(**kwargs)]
-            return self.filter_order_ids(ids)
+            pks = [item.pk for item in self.descendants_raw(**kwargs)]
+            return self.filter_order_pks(pks)
+
+        def descendants_count(self):
+            # ToDo: Implement
+            pass
 
         def self_and_descendants(self, **kwargs):
-            ids = [self.id] + [item.id for item in self.descendants_raw(**kwargs)]
-            return self.filter_order_ids(ids)
+            pks = [self.pk] + [item.pk for item in self.descendants_raw(**kwargs)]
+            return self.filter_order_pks(pks)
 
         def descendants_and_self(self, **kwargs):
-            ids = [item.id for item in self.descendants_raw(**kwargs)] + [self.id]
-            return self.filter_order_ids(ids)
+            pks = [item.pk for item in self.descendants_raw(**kwargs)] + [self.pk]
+            return self.filter_order_pks(pks)
 
         def clan(self, **kwargs):
             """
             Returns a queryset with all ancestors, self, and all descendants
             """
-            ids = (
-                [item.id for item in self.ancestors_raw(**kwargs)]
-                + [self.id]
-                + [item.id for item in self.descendants_raw(**kwargs)]
+            pks = (
+                [item.pk for item in self.ancestors_raw(**kwargs)]
+                + [self.pk]
+                + [item.pk for item in self.descendants_raw(**kwargs)]
             )
-            return self.filter_order_ids(ids)
+            return self.filter_order_pks(pks)
+
+        def clan_count(self):
+            # ToDo: Implement
+            pass
+
+        def siblings(self):
+            # ToDo: Implement
+            pass
+
+        def siblings_count(self):
+            # ToDo: Implement
+            pass
+
+        def self_and_siblings(self):
+            # ToDo: Implement
+            pass
+
+        def siblings_and_self(self):
+            # ToDo: Implement
+            pass
 
         def descendants_edges(self):
             """
@@ -460,21 +550,21 @@ def node_factory(edge_model, children_null=True, base_model=models.Model):
             # ToDo: Implement filters
 
             if self == target_node:
-                return [[self.id]]
+                return [[self.pk]]
 
             downward_clauses, upward_clauses = ("", "")
             query_parameters = {
-                "starting_node": self.id,
-                "ending_node": target_node.id,
+                "starting_node": self.pk,
+                "ending_node": target_node.pk,
                 "max_depth": max_depth,
             }
 
             limiting_fk_nodes_instance = kwargs.get("limiting_fk_nodes_instance", None)
             limiting_fk_edges_instance = kwargs.get("limiting_fk_edges_instance", None)
-            excluded_nodes_queryset = kwargs.get("excluded_nodes_queryset", None)
-            excluded_edges_queryset = kwargs.get("excluded_edges_queryset", None)
-            required_nodes_queryset = kwargs.get("required_nodes_queryset", None)
-            required_edges_queryset = kwargs.get("required_edges_queryset", None)
+            disallowed_nodes_queryset = kwargs.get("disallowed_nodes_queryset", None)
+            disallowed_edges_queryset = kwargs.get("disallowed_edges_queryset", None)
+            allowed_nodes_queryset = kwargs.get("allowed_nodes_queryset", None)
+            allowed_edges_queryset = kwargs.get("allowed_edges_queryset", None)
 
             if limiting_fk_nodes_instance is not None:
                 pass  # Not implemented yet
@@ -484,25 +574,26 @@ def node_factory(edge_model, children_null=True, base_model=models.Model):
                 if fk_field_name is not None:
                     downward_clauses += "\n" + PATH_LIMITING_FK_EDGES_CLAUSE.format(
                         relationship_table=edge_model_table,
+                    pk_name=self.get_pk_name(),
                         fk_field_name=fk_field_name,
                     )
                     query_parameters[
-                        "limiting_fk_edges_instance_id"
-                    ] = limiting_fk_edges_instance.id
+                        "limiting_fk_edges_instance_pk"
+                    ] = limiting_fk_edges_instance.pk
 
-            if excluded_nodes_queryset is not None:
-                downward_clauses += "\n" + EXCLUDED_DOWNWARD_PATH_NODES_CLAUSE
-                query_parameters["excluded_path_node_ids"] = str(
-                    set(excluded_nodes_queryset.values_list("id", flat=True))
+            if disallowed_nodes_queryset is not None:
+                downward_clauses += "\n" + DISALLOWED_DOWNWARD_PATH_NODES_CLAUSE
+                query_parameters["disallowed_path_node_pks"] = str(
+                    set(disallowed_nodes_queryset.values_list("pk", flat=True))
                 )
 
-            if excluded_edges_queryset is not None:
+            if disallowed_edges_queryset is not None:
                 pass  # Not implemented yet
 
-            if required_nodes_queryset is not None:
+            if allowed_nodes_queryset is not None:
                 pass  # Not implemented yet
 
-            if required_edges_queryset is not None:
+            if allowed_edges_queryset is not None:
                 pass  # Not implemented yet
 
             NodeModel = self._meta.model
@@ -510,6 +601,7 @@ def node_factory(edge_model, children_null=True, base_model=models.Model):
             path = NodeModel.objects.raw(
                 DOWNWARD_PATH_QUERY.format(
                     relationship_table=edge_model_table,
+                    pk_name=self.get_pk_name(),
                     downward_clauses=downward_clauses,
                 ),
                 query_parameters,
@@ -527,27 +619,29 @@ def node_factory(edge_model, children_null=True, base_model=models.Model):
                     if "fk_field_name" in locals():
                         upward_clauses += "\n" + PATH_LIMITING_FK_EDGES_CLAUSE.format(
                             relationship_table=edge_model_table,
+                            pk_name=self.get_pk_name(),
                             fk_field_name=fk_field_name,
                         )
 
-                if excluded_nodes_queryset is not None:
-                    upward_clauses += "\n" + EXCLUDED_UPWARD_PATH_NODES_CLAUSE
-                    query_parameters["excluded_path_node_ids"] = str(
-                        set(excluded_nodes_queryset.values_list("id", flat=True))
+                if disallowed_nodes_queryset is not None:
+                    upward_clauses += "\n" + DISALLOWED_UPWARD_PATH_NODES_CLAUSE
+                    query_parameters["disallowed_path_node_pks"] = str(
+                        set(disallowed_nodes_queryset.values_list("pk", flat=True))
                     )
 
-                if excluded_edges_queryset is not None:
+                if disallowed_edges_queryset is not None:
                     pass  # Not implemented yet
 
-                if required_nodes_queryset is not None:
+                if allowed_nodes_queryset is not None:
                     pass  # Not implemented yet
 
-                if required_edges_queryset is not None:
+                if allowed_edges_queryset is not None:
                     pass  # Not implemented yet
 
                 path = NodeModel.objects.raw(
                     UPWARD_PATH_QUERY.format(
                         relationship_table=edge_model_table,
+                        pk_name=self.get_pk_name(),
                         upward_clauses=upward_clauses,
                     ),
                     query_parameters,
@@ -558,8 +652,8 @@ def node_factory(edge_model, children_null=True, base_model=models.Model):
             return path
 
         def path(self, target_node, **kwargs):
-            ids = [item.id for item in self.path_raw(target_node, **kwargs)]
-            return self.filter_order_ids(ids)
+            pks = [item.pk for item in self.path_raw(target_node, **kwargs)]
+            return self.filter_order_pks(pks)
 
         def distance(self, target_node, **kwargs):
             """
@@ -585,6 +679,23 @@ def node_factory(edge_model, children_null=True, base_model=models.Model):
             """
             return bool(not self.children.exists() and not self.parents.exists())
 
+        def is_descendant_of(self, target):
+            # ToDo: Implement
+            pass
+
+        def is_ancestor_of(self, target):
+            # ToDo: Implement
+            pass
+
+        def is_sibling_of(self, target):
+            # ToDo: Implement
+            pass
+
+        def get_node_depth(self):
+            # Depth from furthest root
+            # ToDo: Implement
+            pass
+
         def descendants_tree(self):
             """
             Returns a tree-like structure with descendants
@@ -662,23 +773,23 @@ def node_factory(edge_model, children_null=True, base_model=models.Model):
 
 
 class EdgeManager(models.Manager):
-    def descendants(self, node):
+    def descendants(self, node, **kwargs):
         """
         Returns a queryset of all edges descended from the given node
         """
-        return _filter_order(self.model.objects, "parent", node.self_and_descendants())
+        return _filter_order(self.model.objects, "parent", node.self_and_descendants(**kwargs))
 
-    def ancestors(self, node):
+    def ancestors(self, node, **kwargs):
         """
         Returns a queryset of all edges which are ancestors of the given node
         """
-        return _filter_order(self.model.objects, "child", node.ancestors_and_self())
+        return _filter_order(self.model.objects, "child", node.ancestors_and_self(**kwargs))
 
-    def clan(self, node):
+    def clan(self, node, **kwargs):
         """
         Returns a queryset of all edges for ancestors, self, and descendants
         """
-        return _filter_order(self.model.objects, ["parent", "child"], node.clan())
+        return _filter_order(self.model.objects, ["parent", "child"], node.clan(**kwargs))
 
     def path(self, start_node, end_node, **kwargs):
         """
@@ -690,14 +801,14 @@ class EdgeManager(models.Manager):
             start_node.path(end_node, **kwargs),
         )
 
-    def validate_route(self, edges):
+    def validate_route(self, edges, **kwargs):
         """
         Given a list or set of edges, verify that they result in a contiguous route
         """
         # ToDo: Implement
         pass
 
-    def sort(self, edges):
+    def sort(self, edges, **kwargs):
         """
         Given a list or set of edges, sort them from root-side to leaf-side
         """
@@ -707,15 +818,15 @@ class EdgeManager(models.Manager):
 
 def edge_factory(
     node_model,
-    child_to_field="id",
-    parent_to_field="id",
     concrete=True,
     base_model=models.Model,
 ):
+
     if isinstance(node_model, str):
         try:
             node_model_name = node_model.split(".")[1]
         except IndexError:
+            
             node_model_name = node_model
     else:
         node_model_name = node_model._meta.model_name
@@ -724,13 +835,11 @@ def edge_factory(
         parent = models.ForeignKey(
             node_model,
             related_name=f"{node_model_name}_child",
-            to_field=parent_to_field,
             on_delete=models.CASCADE,
         )
         child = models.ForeignKey(
             node_model,
             related_name=f"{node_model_name}_parent",
-            to_field=child_to_field,
             on_delete=models.CASCADE,
         )
 
@@ -742,6 +851,6 @@ def edge_factory(
         def save(self, *args, **kwargs):
             if not kwargs.pop("disable_circular_check", False):
                 self.parent.__class__.circular_checker(self.parent, self.child)
-            super(Edge, self).save(*args, **kwargs)
+            super().save(*args, **kwargs)
 
     return Edge
